@@ -4,14 +4,17 @@
 # pylint: skip-file
 
 import base64
+import hashlib
 import hmac
 import logging
 import time
 import urllib.parse
 from collections import defaultdict
 from functools import partial, wraps
-from typing import Any, Dict, List, Optional, Union
+import secrets
+from typing import Any, Dict, List, MutableMapping, Optional, Union
 
+from cachetools import TTLCache
 import rest_tools
 import tornado.escape
 import tornado.gen
@@ -370,6 +373,8 @@ class OpenIDLoginHandler(RestHandler, OAuth2Mixin):
 
     Requires the `login_url` application setting to be a full url.
     """
+    _pkcs_challenges: MutableMapping[str, str] = TTLCache(maxsize=10000, ttl=3600)
+
     def initialize(self, oauth_client_id, oauth_client_secret, oauth_client_scope=None, **kwargs):
         super().initialize(**kwargs)
         if not isinstance(self.auth, OpenIDAuth):
@@ -383,24 +388,47 @@ class OpenIDLoginHandler(RestHandler, OAuth2Mixin):
         if oauth_client_scope:
             self.oauth_client_scope = oauth_client_scope.split()
         else:
-            self.oauth_client_scope = ['offline_access', 'profile', 'groups']
+            self.oauth_client_scope = ['profile', 'groups']
+            if oauth_client_secret:
+                self.oauth_client_scope.append('offline_access')
+
+    @classmethod
+    def create_pkce_challenge(cls) -> str:
+        code_verifier = secrets.token_urlsafe(64)
+        code_challenge = base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode('utf-8')).digest()).decode('utf-8').split('=')[0]
+        cls._pkcs_challenges[code_challenge] = code_verifier
+        return code_challenge
+
+    @classmethod
+    def get_pkce_verifier(cls, challenge: str) -> str:
+        if challenge in cls._pkcs_challenges:
+            return cls._pkcs_challenges[challenge]
+        raise KeyError('invalid pkce challenge')
 
     async def get_authenticated_user(
-        self, redirect_uri: str, code: str
+        self, redirect_uri: str, code: str, state: Dict[str, Any]
     ) -> Dict[str, Any]:
         http = self.get_auth_http_client()
-        body = urllib.parse.urlencode({
+        body = {
             'redirect_uri': redirect_uri,
             'code': code,
             'client_id': self.oauth_client_id,
-            'client_secret': self.oauth_client_secret,
             'grant_type': 'authorization_code',
-        })
+        }
+        if self.oauth_client_secret:
+            body['client_secret'] = self.oauth_client_secret
+        else:
+            # use PKCE
+            code_challenge = state.get('code_challenge', None)
+            if not code_challenge:
+                raise tornado.web.HTTPError(400, reason='missing PKCE code challenge')
+            body['code_verifier'] = self.get_pkce_verifier(code_challenge)
+
         response = await http.fetch(
             self._OAUTH_ACCESS_TOKEN_URL,
             method='POST',
             headers={'Content-Type': 'application/x-www-form-urlencoded'},
-            body=body,
+            body=urllib.parse.urlencode(body),
         )
         ret = tornado.escape.json_decode(response.body)
         if not ret.get('id_token', ''):
@@ -446,15 +474,25 @@ class OpenIDLoginHandler(RestHandler, OAuth2Mixin):
             user = await self.get_authenticated_user(
                 redirect_uri=self.get_login_url(),
                 code=self.get_argument('code'),
+                state=data,
             )
+
+            # set expire times (can be 0 in user data, which is an invalid cookie)
+            access_expire = user.get('expires_in', 0)
+            if not access_expire:
+                access_expire = 1800
+            refresh_expire = user.get('refresh_expires_in', 0)
+            if not refresh_expire:
+                refresh_expire = 86400
+
             # Save the user with e.g. set_secure_cookie
             self.set_secure_cookie('access_token', user['access_token'],
-                                   expires_days=float(user['expires_in'])/3600/24)
+                                   expires_days=float(access_expire)/3600/24)
             if 'refresh_token' in user:
                 self.set_secure_cookie('refresh_token', user['refresh_token'],
-                                       expires_days=float(user.get('refresh_expires_in', 86400))/3600/24)
+                                       expires_days=float(refresh_expire)/3600/24)
             self.set_secure_cookie('identity', tornado.escape.json_encode(user['id_token']),
-                                   expires_days=float(user.get('refresh_expires_in', 86400))/3600/24)
+                                   expires_days=float(refresh_expire)/3600/24)
             if data.get('redirect', None):
                 url = data['redirect']
                 if 'state' in data:
@@ -466,15 +504,26 @@ class OpenIDLoginHandler(RestHandler, OAuth2Mixin):
                 raise tornado.web.HTTPError(400, reason='missing redirect')
         else:
             state = {}
-            if self.get_argument('redirect', False):
-                state['redirect'] = self.get_argument('redirect')
+            redirect = self.get_argument('next', None)
+            if not redirect:
+                redirect = self.get_argument('redirect', None)
+            if redirect:
+                state['redirect'] = redirect
             elif not self.settings.get('debug', False):
                 raise tornado.web.HTTPError(400, 'missing redirect')
             if self.get_argument('state', False):
                 state['state'] = self.get_argument('state')
+            extra_params = {}
+            if not self.oauth_client_secret:
+                # use PKCE  https://www.rfc-editor.org/rfc/rfc7636
+                extra_params['code_challenge'] = self.create_pkce_challenge()
+                extra_params['code_challenge_method'] = 'S256'
+                state['code_challenge'] = extra_params['code_challenge']
+            extra_params["state"] = self._encode_state(state)
+
             self.authorize_redirect(
                 redirect_uri=self.get_login_url(),
                 client_id=self.oauth_client_id,
                 scope=self.oauth_client_scope,
-                extra_params={"state": self._encode_state(state)},
+                extra_params=extra_params,
                 response_type='code')
